"""
/***************************************************************************
 qgSurf - plugin for Quantum GIS

 Processing of geological planes and surfaces

                              -------------------
        begin                : 2011-12-21
        copyright            : (C) 2011-2022 by Mauro Alberti
        email                : alberti.m65@gmail.com

 ***************************************************************************/

# licensed under the terms of GNU GPL 3

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from PyQt5.QtWidgets import *

from pygsf.orientations.orientations import Plane as GPlane
from pygsf.geometries.planes import *

from gst.qgis.orientations import *
from gst.qgis.canvas import *
from gst.qgis.tables import *
from gst.qgis.rasters import *
from gst.qgis.messages import *
from gst.io.vectors.ogr_io import *

from ..configurations.general_params import *
from ..configurations.plane_points_distances_params import *

from ..ui.auxiliary_windows.distances import DistancesSrcPtLyrDlg, tFieldUndefined


def get_distances_input_params(dialog):

    def parse_field_choice(val, choose_message):

        if val == choose_message:
            return None
        else:
            return val

    plyr = dialog.point_layer

    plyr_id_name_field = parse_field_choice(
        dialog.cmbIdSrcFld.currentText(),
        tFieldUndefined)
    if not plyr_id_name_field:
        return False, "Incorrect id field choice"

    plyr_x_name_field = parse_field_choice(
        dialog.cmbXSrcFld.currentText(),
        tFieldUndefined)
    if not plyr_x_name_field:
        return False, "Incorrect x field choice"

    plyr_y_name_field = parse_field_choice(
        dialog.cmbYSrcFld.currentText(),
        tFieldUndefined)
    if not plyr_y_name_field:
        return False, "Incorrect y field choice"

    plyr_z_name_field = parse_field_choice(
        dialog.cmbZSrcFld.currentText(),
        tFieldUndefined)
    if not plyr_z_name_field:
        return False, "Incorrect z field choice"

    # geological plane parameters section

    gplane_dipdir = dialog.spnTargetAttDipDir.value()
    gplane_dipangle = dialog.spnTargetAttDipAng.value()

    try:

        gplane_srcpt_x_coord = float(dialog.qlndtXPlaneSrcFld.text())
        gplane_srcpt_y_coord = float(dialog.qlndtYPlaneSrcFld.text())
        gplane_srcpt_z_coord = float(dialog.qlndtZPlaneSrcFld.text())

    except:

        return False, "Incorrect source point coordinates"

    # output shapefile

    output_shapefile_path = dialog.lnedtOutFilename.text()
    if not output_shapefile_path:
        return False, "Missing output shapefile path"

    # returning parameters

    return True, dict(
        point_layer=plyr,
        point_layer_flds=(plyr_id_name_field, plyr_x_name_field, plyr_y_name_field, plyr_z_name_field),
        gplane_params=dict(
            pl_attitude=(gplane_dipdir, gplane_dipangle),
            src_pt=(gplane_srcpt_x_coord, gplane_srcpt_y_coord, gplane_srcpt_z_coord)),
        output_shapefile_path=output_shapefile_path)


class PtsPlnDistancesWidget(QWidget):

    sgnWindowClosed = pyqtSignal()

    def __init__(self,
                 canvas,
                 plugin_name,
                 plugin_folder
                 ):

        super(PtsPlnDistancesWidget, self).__init__()
        self.mapCanvas = canvas
        self.plugin_name = plugin_name
        self.plugin_folder = plugin_folder

        self.point_layer = None
        self.point_layer_fields = []
        self.distances_analysis_params = None

        self.setup_gui()

    def setup_gui(self): 

        self.layout = QVBoxLayout()
        
        self.layout.addWidget(self.setup_inputdata())
        self.layout.addWidget(self.setup_processing())
        self.layout.addWidget(self.setup_help())

        self.setLayout(self.layout)
        self.setWindowTitle("{} - points distances to geological plane".format(self.plugin_name))
        self.adjustSize()

    def setup_inputdata(self):
        
        grpInput = QGroupBox("Define params")
        
        layout = QVBoxLayout() 
        
        self.pshDefinePointLayer = QPushButton(self.tr("I/O params"))
        self.pshDefinePointLayer.clicked.connect(self.user_define_distances_inparams)
        layout.addWidget(self.pshDefinePointLayer)
        
        grpInput.setLayout(layout)
        
        return grpInput

    def setup_processing(self):
        
        grpProcessing = QGroupBox("Processing")
        
        layout = QGridLayout()

        self.pshCalculateAngles = QPushButton(self.tr("Calculate"))
        self.pshCalculateAngles.clicked.connect(self.calculate_distances)
        layout.addWidget(self.pshCalculateAngles, 0, 0, 1, 1)

        grpProcessing.setLayout(layout)
        
        return grpProcessing

    def setup_help(self):

        group_box = QGroupBox("Help")

        layout = QVBoxLayout()

        self.pshHelp = QPushButton(self.tr("Open help"))
        self.pshHelp.clicked.connect(self.open_help)
        layout.addWidget(self.pshHelp)

        group_box.setLayout(layout)

        return group_box

    def user_define_distances_inparams(self):

        self.distances_analysis_params = None

        if len(loaded_point_layers()) == 0:
            warn_qgis(
                self.plugin_name,
                "No available point layers"
            )
            return

        dialog = DistancesSrcPtLyrDlg()
        if dialog.exec_():
            success, cntnt = get_distances_input_params(dialog)
        else:
            warn_qgis(
                self.plugin_name,
                "No input data defined"
            )
            return

        if not success:
            warn_qgis(
                self.plugin_name,
                cntnt
            )
            return
        else:
            structural_input_params = cntnt

        self.point_layer = structural_input_params["point_layer"]
        self.point_layer_fields = structural_input_params["point_layer_flds"]
        self.gPlaneAttitude = structural_input_params["gplane_params"]["pl_attitude"]
        self.gPlaneSrcPt = structural_input_params["gplane_params"]["src_pt"]
        self.outputShapefilePath = structural_input_params["output_shapefile_path"]

        info_qgis(
            self.plugin_name,
            f"Now results can be calculated"
        )

    def calculate_distances(self):

        if not self.point_layer or \
           not self.point_layer_fields:
            warn_qgis(
                self.plugin_name,
                "Input data are not defined"
            )
            return

        # get input point layer data

        point_layer_data = lyr_attrs(
            layer=self.point_layer,
            fields=self.point_layer_fields)

        len_data = len(point_layer_data)

        if len_data == 0:
            warn_qgis(
                self.plugin_name,
                "No data to calculate"
            )
            return

        cartplane_srcpt = Point(*self.gPlaneSrcPt)
        crs_projection = projectCrs()

        geoplane_dipdir, geoplane_dipangle = self.gPlaneAttitude

        azimuth_correction = calculate_azimuth_correction(
            src_pt=cartplane_srcpt,
            crs=crs_projection)
        print("Azimuth correction: {}".format(azimuth_correction))

        geoplane_dipdir_corrected = (geoplane_dipdir + azimuth_correction) % 360.0
        prj_corrected_geolplane = GPlane(geoplane_dipdir_corrected, geoplane_dipangle)

        #cartplane = prj_corrected_geolplane.toCPlane(cartplane_srcpt)
        cartplane = CPlane3D.from_geol_plane(prj_corrected_geolplane, cartplane_srcpt)

        ltAttributes = []
        ltfGeometries = []

        for geog_pt_data in point_layer_data:

            rec_id, geog_x, geog_y, geog_z = geog_pt_data
            geog_pt = Point(geog_x, geog_y, geog_z)

            prj_pt, err = cartplane.project_point(geog_pt)
            if err:
                warn_qgis(
                    self.plugin_name,
                    f"{repr(err)}"
                )

            prj_x, prj_y, prj_z = prj_pt.xyz()

            geo_prj_ds = cartplane.distance_to_point(geog_pt)

            src_geog_vect = Segment(
                start_pt=cartplane_srcpt,
                end_pt=geog_pt).as_vector2d()

            src_prj_vect = Segment(
                start_pt=cartplane_srcpt,
                end_pt=prj_pt).as_vector2d()

            src_geo_ds = src_geog_vect.length
            src_prj_ds = src_prj_vect.length

            src_geo_az = src_geog_vect.azimuth_degr()
            src_prj_az = src_prj_vect.azimuth_degr()

            prj_geo_an = src_geog_vect.angle_with(src_prj_vect)

            vals = (
                rec_id,
                geog_x,
                geog_y,
                geog_z,
                prj_x,
                prj_y,
                prj_z,
                geo_prj_ds,
                src_geo_ds,
                src_geo_az,
                src_prj_ds,
                src_prj_az,
                prj_geo_an
            )

            ltAttributes.append(vals)
            ltfGeometries.append((geog_x, geog_y, geog_z))

        ltFieldsNames = [field_dict["name"] for field_dict in output_table_fields_pars]

        out_shape_pth = self.outputShapefilePath

        shapefile_datasource, point_shapelayer = shapefile_create(
            path=out_shape_pth,
            geom_type=ogr.wkbPoint25D,
            fields_dict_list=output_table_fields_pars)

        print(f"Shapefile {out_shape_pth} created")

        if not shapefile_datasource or not point_shapelayer:
            warn_qgis(
                self.plugin_name,
                f"Unable to create output shapefile {out_shape_pth}"
            )
            return

        success, msg = try_write_pt_shapefile(
            point_layer=point_shapelayer,
            geoms=ltfGeometries,
            field_names=ltFieldsNames,
            attrs=ltAttributes)

        """
        try:
            #del point_shapelayer
            del shapefile_datasource
        except:
            pass
        """

        if success:
            info_qgis(
                self.plugin_name,
                f"Results written in output shapefile {out_shape_pth}"
            )
        else:
            warn_qgis(
                self.plugin_name,
                msg
            )

    def closeEvent(self, event):

        settings = QSettings(settings_name, plugin_nm)

        settings.setValue("PtsPlnDistancesWidget/size", self.size())
        settings.setValue("PtsPlnDistancesWidget/position", self.pos())

        self.sgnWindowClosed.emit()

    def open_help(self):

        dialog = HelpDialog(
            self.plugin_name,
            self.plugin_folder
        )
        dialog.exec_()


class HelpDialog(QDialog):

    def __init__(self,
                 plugin_name,
                 plugin_folder,
                 parent=None):

        super(HelpDialog, self).__init__(parent)

        layout = QVBoxLayout()

        # About section

        helpTextBrwsr = QTextBrowser(self)

        url_path = f"file:///{plugin_folder}/help/help_pts_plane_dists.html"
        helpTextBrwsr.setSource(QUrl(url_path))
        helpTextBrwsr.setSearchPaths([f'{plugin_folder}/help/images/distances'])
        helpTextBrwsr.setMinimumSize(700, 600)

        layout.addWidget(helpTextBrwsr)

        self.setLayout(layout)

        self.setWindowTitle("{} - Points-plane distances help".format(plugin_name))

